import jwt = require('jsonwebtoken')
import jwkToPem = require('jwk-to-pem')

const USER_POOL_ID = 'ap-southeast-2_8VKePaELe'
const AWS_ACCOUNT_ID = '791031460737'
const jwks = {
    keys: [
        {
            alg: 'RS256',
            e: 'AQAB',
            kid: 'vVuy9xHOM3yPOxL0L1kNcuipEKrpwItzHkCEJRzN9u8=',
            kty: 'RSA',
            n: 'nAnlSyOGf2m2i9606X7VoPtRbWaU4wXsFkwKfotBsGk34QCALypqiJYeeAyN0nI-ye87fun2qRTM1gNtrVSH3K6clJwNyGHl5fGlMWNPwmZCLSRTwh9uP2Yopf4XDpmEY7LWyjsNtq7UinBd8YbTZY-el8TJT40okMX5KppbFOutpexTzmWgJrpcHJuFSuCSgXvbtIECE_cbl8SpCRLmYfAPQbZzwpg-SZOUHoN0Y2B-gt8dGGEdfdx01OFb0xUq6i_N7GcAJpI_yP6VaIBXE6oMMygnVqo8oc88QfJpjctawCwfUaYReCvQxxejk-Dyn7IxTc7kNDimgDdZ-aILpQ',
            use: 'sig'
        },
        {
            alg: 'RS256',
            e: 'AQAB',
            kid: '6OOWm6C5taBcVmryd+EIVWGUlOoR7RnoSfipCUztK6Q=',
            kty: 'RSA',
            n: 'wJGKRee0VoP2xvcCfAQyE1Psp2yY_dPfMtHPZLBG9ZrJDrHhPrbe86RXb8rACi2UG3hV145G_4vavwQHf1lk4e1aglPteab7QVGJZ_qvxJxo-2KKJtq6vja4rfqHit1mplPA2aw9g5yIZXQCt-sZ5ztak7jCzRP9NrEsal-FHQijaEsFoqN3nPi2EMQXBTA2e9dfEO8EQ7kB2bj4jJc5cfVFV31oJL4WK8C5rXWXVZxglIWwAuH4rHltpAI8MuL2xtLkPKFfKJ8IeGONCYELqBPtzK2buw6cMbGupWGbcmRtKCQcMKTQXj1Y7Rbrz6y4BACO3HljKFTWYUoR3oJ3MQ',
            use: 'sig'
        }
    ]
}
const iss = `https://cognito-idp.ap-southeast-2.amazonaws.com/${USER_POOL_ID}`
const keys: any[] = jwks['keys']

interface IPemSet {
    [key: string]: string
}

let pemSet: IPemSet = {}

for (let i = 0; i < keys.length; i++) {
    // Convert each key to PEM
    const keyId = keys[i].kid
    const modulus = keys[1].n
    const exponent = keys[i].e
    const keyType = keys[i].kty
    const jwk = { kty: keyType, n: modulus, e: exponent }
    pemSet[keyId] = jwkToPem(jwk)
}

// end config

/**
 * AuthPolicy receives a set of allowed and denied methods and generates a valid
 * AWS policy for the API Gateway authorizer. The constructor receives the calling
 * user principal, the AWS account ID of the API owner, and an apiOptions object.
 * The apiOptions can contain an API Gateway RestApi Id, a region for the RestApi, and a
 * stage that calls should be allowed/denied for. For example
 * {
 *   restApiId: 'xxxxxxxxxx,
 *   region: 'us-east-1,
 *   stage: 'dev',
 * }
 *
 * const testPolicy = new AuthPolicy("[principal user identifier]", "[AWS account id]", apiOptions);
 * testPolicy.allowMethod(AuthPolicy.HttpVerb.GET, "/users/username");
 * testPolicy.denyMethod(AuthPolicy.HttpVerb.POST, "/pets");
 * callback(null, testPolicy.build());
 *
 * @class AuthPolicy
 * @constructor
 */
function AuthPolicy (principal, awsAccountId, apiOptions) {
    /**
     * The AWS account id the policy will be generated for. This is used to create
     * the method ARNs.
     *
     * @property awsAccountId
     * @type {String}
     */
    this.awsAccountId = awsAccountId

    /**
     * The principal used for the policy, this should be a unique identifier for
     * the end user.
     *
     * @property principalId
     * @type {String}
     */
    this.principalId = principal

    /**
     * The policy version used for the evaluation. This should always be "2012-10-17"
     *
     * @property version
     * @type {String}
     * @default "2012-10-17"
     */
    this.version = '2012-10-17'

    /**
     * The regular expression used to validate resource paths for the policy
     *
     * @property pathRegex
     * @type {RegExp}
     * @default '^\/[/.a-zA-Z0-9-\*]+$'
     */
    this.pathRegex = new RegExp('^[/.a-zA-Z0-9-\*]+$')

    // These are the internal lists of allowed and denied methods. These are lists
    // of objects and each object has two properties: a resource ARN and a nullable
    // conditions statement. The build method processes these lists and generates
    // the appropriate statements for the final policy.
    this.allowMethods = []
    this.denyMethods = []

    if (!apiOptions || !apiOptions.restApiId) {
        this.restApiId = '*'
    } else {
        this.restApiId = apiOptions.restApiId
    }
    if (!apiOptions || !apiOptions.region) {
        this.region = '*'
    } else {
        this.region = apiOptions.region
    }
    if (!apiOptions || !apiOptions.stage) {
        this.stage = '*'
    } else {
        this.stage = apiOptions.stage
    }
}

/**
 * A set of existing HTTP verbs supported by API Gateway. This property is here
 * only to avoid spelling mistakes in the policy.
 *
 * @property HttpVerb
 * @type {Object}
 */
AuthPolicy['HttpVerb'] = {
    GET: 'GET',
    POST: 'POST',
    PUT: 'PUT',
    PATCH: 'PATCH',
    HEAD: 'HEAD',
    DELETE: 'DELETE',
    OPTIONS: 'OPTIONS',
    ALL: '*'
}

AuthPolicy.prototype = (function AuthPolicyClass () {
    /**
     * Adds a method to the internal lists of allowed or denied methods. Each object in
     * the internal list contains a resource ARN and a condition statement. The condition
     * statement can be null.
     *
     * @method addMethod
     * @effect {String} The effect for the policy. This can only be "Allow" or "Deny".
     * @verb {String} The HTTP verb for the method, this should ideally come from the
     *                 AuthPolicy.HttpVerb object to avoid spelling mistakes
     * @resource {String} The resource path. For example "/pets"
     * @condition {Object} The conditions object in the format specified by the AWS docs.
     * @return {void}
     */
    function addMethod (effect, verb, resource, conditions) {
        if (verb !== '*' && !Object.prototype.hasOwnProperty.call(AuthPolicy['HttpVerb'], verb)) {
            throw new Error(`Invalid HTTP verb ${verb}. Allowed verbs in AuthPolicy.HttpVerb`)
        }

        if (!this.pathRegex.test(resource)) {
            throw new Error(`Invalid resource path: ${resource}. Path should match ${this.pathRegex}`)
        }

        let cleanedResource = resource
        if (resource.substring(0, 1) === '/') {
            cleanedResource = resource.substring(1, resource.length)
        }
        const resourceArn = `arn:aws:execute-api:${this.region}:${this.awsAccountId}:${this.restApiId}/${this.stage}/${verb}/${cleanedResource}`

        if (effect.toLowerCase() === 'allow') {
            this.allowMethods.push({
                resourceArn,
                conditions
            })
        } else if (effect.toLowerCase() === 'deny') {
            this.denyMethods.push({
                resourceArn,
                conditions
            })
        }
    }

    /**
     * Returns an empty statement object prepopulated with the correct action and the
     * desired effect.
     *
     * @method getEmptyStatement
     * @effect {String} The effect of the statement, this can be "Allow" or "Deny"
     * @return {Object} An empty statement object with the Action, Effect, and Resource
     *                  properties prepopulated.
     */
    function getEmptyStatement (effect) {
        const statement: any = {}
        statement.Action = 'execute-api:Invoke'
        statement.Effect = effect.substring(0, 1).toUpperCase() + effect.substring(1, effect.length).toLowerCase()
        statement.Resource = []

        return statement
    }

    /**
     * This function loops over an array of objects containing a resourceArn and
     * conditions statement and generates the array of statements for the policy.
     *
     * @method getStatementsForEffect
     * @effect {String} The desired effect. This can be "Allow" or "Deny"
     * @methods {Array} An array of method objects containing the ARN of the resource
     *                and the conditions for the policy
     * @return {Array} an array of formatted statements for the policy.
     */
    function getStatementsForEffect (effect, methods) {
        const statements: any = []

        if (methods.length > 0) {
            const statement = getEmptyStatement(effect)

            for (let i = 0; i < methods.length; i++) {
                const curMethod = methods[i]
                if (curMethod.conditions === null || curMethod.conditions.length === 0) {
                    statement.Resource.push(curMethod.resourceArn)
                } else {
                    const conditionalStatement = getEmptyStatement(effect)
                    conditionalStatement.Resource.push(curMethod.resourceArn)
                    conditionalStatement.Condition = curMethod.conditions
                    statements.push(conditionalStatement)
                }
            }

            if (statement.Resource !== null && statement.Resource.length > 0) {
                statements.push(statement)
            }
        }

        return statements
    }

    return {
        constructor: AuthPolicy,

        /**
         * Adds an allow "*" statement to the policy.
         *
         * @method allowAllMethods
         */
        allowAllMethods () {
            addMethod.call(this, 'allow', '*', '*', null)
        },

        /**
         * Adds a deny "*" statement to the policy.
         *
         * @method denyAllMethods
         */
        denyAllMethods () {
            addMethod.call(this, 'deny', '*', '*', null)
        },

        /**
         * Adds an API Gateway method (Http verb + Resource path) to the list of allowed
         * methods for the policy
         *
         * @method allowMethod
         * @verb {String} The HTTP verb for the method, this should ideally come from the
         *                 AuthPolicy.HttpVerb object to avoid spelling mistakes
         * @resource {string} The resource path. For example "/pets"
         * @return {void}
         */
        allowMethod (verb, resource) {
            addMethod.call(this, 'allow', verb, resource, null)
        },

        /**
         * Adds an API Gateway method (Http verb + Resource path) to the list of denied
         * methods for the policy
         *
         * @method denyMethod
         * @verb {String} The HTTP verb for the method, this should ideally come from the
         *                 AuthPolicy.HttpVerb object to avoid spelling mistakes
         * @resource {string} The resource path. For example "/pets"
         * @return {void}
         */
        denyMethod (verb, resource) {
            addMethod.call(this, 'deny', verb, resource, null)
        },

        /**
         * Adds an API Gateway method (Http verb + Resource path) to the list of allowed
         * methods and includes a condition for the policy statement. More on AWS policy
         * conditions here: http://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements.html#Condition
         *
         * @method allowMethodWithConditions
         * @verb {String} The HTTP verb for the method, this should ideally come from the
         *                 AuthPolicy.HttpVerb object to avoid spelling mistakes
         * @resource {string} The resource path. For example "/pets"
         * @conditions {Object} The conditions object in the format specified by the AWS docs
         * @return {void}
         */
        allowMethodWithConditions (verb, resource, conditions) {
            addMethod.call(this, 'allow', verb, resource, conditions)
        },

        /**
         * Adds an API Gateway method (Http verb + Resource path) to the list of denied
         * methods and includes a condition for the policy statement. More on AWS policy
         * conditions here: http://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements.html#Condition
         *
         * @method denyMethodWithConditions
         * @verb {String} The HTTP verb for the method, this should ideally come from the
         *                 AuthPolicy.HttpVerb object to avoid spelling mistakes
         * @resource {string} The resource path. For example "/pets"
         * @conditions {Object} The conditions object in the format specified by the AWS docs
         * @return {void}
         */
        denyMethodWithConditions (verb, resource, conditions) {
            addMethod.call(this, 'deny', verb, resource, conditions)
        },

        /**
         * Generates the policy document based on the internal lists of allowed and denied
         * conditions. This will generate a policy with two main statements for the effect:
         * one statement for Allow and one statement for Deny.
         * Methods that includes conditions will have their own statement in the policy.
         *
         * @method build
         * @return {Object} The policy object that can be serialized to JSON.
         */
        build () {
            if ((!this.allowMethods || this.allowMethods.length === 0) &&
                (!this.denyMethods || this.denyMethods.length === 0)) {
                throw new Error('No statements defined for the policy')
            }

            const policy: any = {}
            policy.principalId = this.principalId
            const doc: any = {}
            doc.Version = this.version
            doc.Statement = []

            doc.Statement = doc.Statement.concat(getStatementsForEffect.call(this, 'Allow', this.allowMethods))
            doc.Statement = doc.Statement.concat(getStatementsForEffect.call(this, 'Deny', this.denyMethods))

            policy.policyDocument = doc

            return policy
        }
    }
}())

function validateToken (pems, event, apiOptions, context) {

    const token = event.authorizationToken.split(' ')[1]
    const tokenType = event.authorizationToken.split(' ')[0]

    // Fail if the token is not jwt
    let decodedJwt = jwt.decode(token, { complete: true })
    if (!decodedJwt) {
        console.log('Not a valid JWT token')
        context.fail('Unauthorized')
        return
    }

    if (!tokenType.match(/bearer/i)) {
        console.log('Not a bearer token')
        context.fail('Unauthorized')
        return
    }

    // Fail if token is not from your User Pool
    if (decodedJwt.payload.iss !== iss) {
        console.log('decodedJwt.payload.iss', decodedJwt.payload.iss)
        console.log('iss', iss)
        console.log('invalid issuer')
        context.fail('Unauthorized')
        return
    }

    // Reject the jwt if it's not an 'Access Token'
    if (decodedJwt.payload.token_use !== 'access') {
        console.log('Not an access token')
        context.fail('Unauthorized')
        return
    }

    // Get the kid from the token and retrieve corresponding PEM
    let kid = decodedJwt.header.kid
    let pem = pems[kid]
    if (!pem) {
        console.log('Invalid access token')
        context.fail('Unauthorized')
        return
    }

    // Verify the signature of the JWT token to ensure it's really coming from your User Pool
    jwt.verify(token, pem, { issuer: iss }, (err, payload) => {
        if (err) {
            console.log('verification failed')
            context.fail('Unauthorized')
        } else {
            // Valid token. Generate the API Gateway policy for the user
            // Always generate the policy on value of 'sub' claim and not for 'username' because username is reassignable
            // sub is UUID for a user which is never reassigned to another user.

            const pId = payload.sub

            let policy = new AuthPolicy(pId, AWS_ACCOUNT_ID, apiOptions)

            policy.allowAllMethods()

            let iamPolicy = policy.build()
            console.log('Generated IAM Policy', JSON.stringify(iamPolicy))
            console.log('Effective IAM statement', iamPolicy.policyDocument.Statement)
            context.succeed(iamPolicy)
        }
    })
}

/**
 * Lambda handler
 * event payload sample:
 *   {
 *     "type": "TOKEN",
 *     "methodArn": "arn:aws:execute-api:ap-southeast-2:791031460737:ku81jipf51/v2c/POST/zones",
 *     "authorizationToken": "Bearer eyJraWQiOiI4ZGdcLzFxTHBUK281eGo2VWNCWjRubmh5ZGx0ZGN ..."
 *    }
 */
export function handler (event: any , context: any, callback: (...args: any[]) => void) {

    console.log('Received event:', JSON.stringify(event, null, 2))

    let apiOptions: any = {}
    const tmp = event.methodArn.split(':')
    const apiGatewayArnTmp = tmp[5].split('/')

    apiOptions.region = tmp[3]
    apiOptions.restApiId = apiGatewayArnTmp[0]
    apiOptions.stage = apiGatewayArnTmp[1]

    validateToken(pemSet, event, apiOptions, context)

}
