const { ResponseHandler } = require('../utils/ResponseHandler')
const AWS = require('aws-sdk')
const dynamodb = new AWS.DynamoDB({maxRetries: 5, retryDelayOptions: {base: 300} })
AWS.config.update({region: "ap-southeast-2"});
var docClient = new AWS.DynamoDB.DocumentClient({service : dynamodb})

const { TABLE_NAME, SUB_INDEX_NAME, DYNAMO_FAIL } = require('../constants')

exports.addUser = (user) => {
    console.log('Dynamo: addUser called')

    // Don't store the password in dynamo
    delete user.password;
    // userId stored as 'sub'
    delete user.userId;

    return new Promise((resolve, reject) => {
        var params = {
            TableName: TABLE_NAME,
            Item: user
        }

        docClient.put(params, function(err, data) {
           if (err) {
               console.error("Unable to add user", ". Error JSON:", JSON.stringify(err, null, 2))
               reject({
                   operation: DYNAMO_FAIL,
                   data
               })
           } else {
               resolve(user.sub)
           }
        })
    })
}

exports.getUsersByAccountId = (accountId) => {
    console.log('getUsersByAccountId called')

    return new Promise((resolve, reject) => {
        var params = {
            TableName: TABLE_NAME,
            KeyConditionExpression: "#acctId = :id",
            ExpressionAttributeNames:{
                "#acctId": "accountId"
            },
            ExpressionAttributeValues: {
                ":id": accountId
            }
        };

        docClient.query(params, function(err, data) {
            if (err) {
                reject(err)
            } else {
                resolve(data)
            }
        });
    })
}

exports.getUser = (userId) => {
    console.log('dynamo: getUser called')

    return new Promise((resolve, reject) => {
        var params = {
            TableName: TABLE_NAME,
            IndexName: SUB_INDEX_NAME,
            ExpressionAttributeNames: {
                "#userId": "sub"
            },
            KeyConditionExpression: "#userId = :id",
            ExpressionAttributeValues: {
                ":id": userId
            }
        };

        docClient.query(params, function(err, data) {
            if (err) {
                reject(err)
            } else if (data.Count === 0) {
                reject({statusCode: 404, message: `${userId} not found in ${TABLE_NAME} dynamo table`})
            } else {
                resolve(data.Items[0])
            }
        });
    })
}

exports.deleteUser = (user) => {
    console.log('dynamo: deleteUser called')

    return new Promise((resolve, reject) => {
        const { accountId, sub, email } = user

        var params = {
            TableName: TABLE_NAME,
            Key: { accountId, sub },
            ConditionExpression: "email = :email",
            ExpressionAttributeValues: {":email": email},
        };

        docClient.delete(params, function(err, data) {
            if (err) {
                reject(err)
            } else {
                console.log('resolvin', data)
                resolve(data)
            }
        })
    })
}

// Will add the patch endpoint after the first release of the API.
// Need to make a decision about whether we use RFC 7396 or RFC 5789
// exports.patchUser = (accountId, userId, props) => {
//     console.log('patchUser called')
//
//     return new Promise((resolve, reject) => {
//         var params = {
//             TableName: TABLE_NAME,
//             Key:{
//                 accountId,
//                 sub: userId
//             },
//             UpdateExpression: "set firstName = :fn, lastName=:ln",
//             ExpressionAttributeValues: {
//                 ":fn": props.firstName,
//                 ":ln": props.lastName,
//             },
//         };
//
//         docClient.update(params, function(err, data) {
//             if (err) {
//                 reject(err)
//             } else {
//                 console.log('data', data)
//                 resolve(data)
//             }
//         });
//     })
// }
