const mongoose = require("mongoose")
mongoose.Promise = global.Promise
const User = require('@bluedot-innovation/models').User(mongoose)
const UserSchema = require('@bluedot-innovation/models').Schemas.User

export const validateUserDocument = userData => {
    const sanitizedData = removeExtraProperties(UserSchema, userData)
    const userDocument = new User(sanitizedData)

    const userDocValidation = new Promise((resolve, reject) => {
        userDocument.validate(error => {
            if (error) {
                reject(error)
            } else {
                resolve()
            }
        })
    })

    return userDocValidation
}

/**
 * removeExtraProperties(schema, data)
 * removes any properties from the data that aren't in the schema
 */
function removeExtraProperties(schema, data) {
    const schemaObject = schema.obj

    for (const prop in data) {
        if (schemaObject[prop] === undefined) {
            delete data[prop]
        }
    }

    return data
}
