interface IResponse {
    statusCode: number,
    headers?: string,
    errors?: any,
    message?: any,
    body?: string
}

export class ResponseHandler {
    public static generateSuccessResponse (statusCode, message?, headerObj?) {
        let response: IResponse = {
            statusCode: statusCode || 200,
            headers: headerObj || { 'x-custom-header': 'custom header value' }
        }

        if (message) response.body = message

        return response
    }

    public static generateErrorResponseObject (statusCode, message?, headerObj?) {
        let response: IResponse = {
            statusCode: statusCode || 400,
            headers: headerObj || { 'x-custom-header': 'custom header value' }
        }
        if (message && (typeof message) === 'object' && !!message.name && (!!message._message)) {
            // MongooseError
            message.error_summary = message._message
            response.errors = message
        } else {
            // JsonPatchError
            if (typeof message === 'object' && !!message.message && message.message.op) {
                response.errors = {
                    message: `${message.message}. Operation: ${message.operation.op}, Path: ${message.operation.path}`,
                    error_summary: 'API Validation Error: PATCH'
                }
            } else if (typeof message === 'object' && message.message) {
                response.errors = {
                    message: `${message.message}`,
                    error_summary: message.code
                }
            } else {
                // Self generated
                response.errors = {
                    message: message || 'Generic error - bad request',
                    error_summary: message
                }
            }

        }

        return new Error(JSON.stringify(response))
    }
}
